from pyhocon import ConfigTree
from pyhocon.exceptions import ConfigMissingException

from .parameters import Parameter, Wrapper


class Configurable:
    @classmethod
    def __init_subclass__(cls, path='', adopt=True):
        if path:
            for par in cls.get_parameters():
                par.name = '{}.{}'.format(path, par.name)
        if adopt:
            adopted = []
            for base in filter(lambda x: issubclass(x, Configurable), cls.__bases__):
                adopted.extend(base.get_parameters())
            for par in filter(lambda x: not hasattr(cls, x.field_name), adopted):
                setattr(cls, par.field_name, par)

    def __init__(self, configuration=None):
        def load_parameter(par):
            if par.name in ('', '.'):
                return par.load(configuration)
            try:
                specification = configuration[par.name]
            except ConfigMissingException as err:
                if par.optional:
                    return par.default
                else:
                    raise err
            else:
                return par.load(specification)

        if configuration is None:
            configuration = ConfigTree()  # To prevent TypeError later on, when subscripting.

        for parameter in self.get_parameters():
            setattr(self, parameter.field_name, load_parameter(parameter))

    @classmethod
    def get_parameters(cls):
        return filter(
            lambda y: isinstance(y, (Parameter, Wrapper)),
            map(
                lambda x: getattr(cls, x),
                dir(cls)
            )
        )



from collections import OrderedDict
import copy
from functools import partial, reduce
import re
from typing import Any, Union

from pyhocon import ConfigTree
from pyhocon.exceptions import ConfigMissingException
try:
    import numpy as np
except ImportError:
    np = None
try:
    from scipy.constants import physical_constants
except ImportError:
    physical_constants = None

from hanna.physics import units
from hanna.utils import safe_math_eval


class Parameter:
    def __or__(self, parameter):
        return self.fallback(parameter)

    def __rmul__(self, times):
        vector = Vector(type(self), n=times, as_=tuple)
        vector.template = self
        return vector

    def __init__(self, name=None, default=None, optional=False, info=None, choices=None):
        self.name = name
        self._default = default
        self.optional = default is not None or optional
        self.info = info
        self.choices = choices or []
        self._field_name = None
        self.constraints = [Constraint(lambda x: not self.choices or x in self.choices,
                                       lambda x: f'Illegal choice ({x} not in {self.choices})')]
        self.transformations = []

    def __set_name__(self, owner, name):
        self._field_name = name
        name = name.lstrip('_')
        if self.name is None:
            self.name = name
        elif self.name.endswith('.'):
            self.name += name

    def __repr__(self):
        return (
            f'{self.__class__.__name__}(name={repr(self.name)}, default={repr(self._default)}, '
            f'optional={repr(self.optional)}, info={repr(self.info)}, choices={self.choices or None})'
        )

    def __eq__(self, val):
        self.constraints.append(Constraint(lambda x: x == val, f'Not equal ({{value}} ≠ {val})'))
        return self

    def __ge__(self, lim):
        self.constraints.append(Constraint(lambda x: x >= lim, f'Too small ({{value}} < {lim})'))
        return self

    def __gt__(self, lim):
        self.constraints.append(Constraint(lambda x: x > lim, f'Too small ({{value}} ≤ {lim})'))
        return self

    def __le__(self, lim):
        self.constraints.append(Constraint(lambda x: x <= lim, f'Too large ({{value}} > {lim})'))
        return self

    def __lt__(self, lim):
        self.constraints.append(Constraint(lambda x: x < lim, f'Too large ({{value}} ≥ {lim})'))
        return self

    def __ne__(self, val):
        self.constraints.append(Constraint(lambda x: x != val, f'Must not be equal ({{value}} = {val})'))
        return self

    def __rshift__(self, transformation):
        try:
            return self.transform(*transformation)
        except TypeError:
            return self.transform(transformation)

    @property
    def abs(self):
        return Characteristic(self, abs)

    @property
    def len(self):
        return Characteristic(self, len)

    @property
    def field_name(self):
        return self._field_name or self.name.rsplit('.', 1)[-1].lower()

    @property
    def default(self):
        if self._default is None and not self.optional:
            raise AttributeError
        return self._default

    @default.setter
    def default(self, value):
        self._default = value
        self.optional = True

    def constrain(self, constraint, error_msg=''):
        self.constraints.append(Constraint(constraint, error_msg))
        return self

    def fallback(self, parameter):
        return FallbackGroup(self, [parameter])

    def load(self, value: Union[str, list, dict, ConfigTree]) -> Any:
        result = reduce(lambda x, t: t(x), self.transformations, value)
        for constraint in self.constraints:
            try:
                constraint(result)
            except ValueError as err:
                raise ValueError(f'"{self.name}" {str(err)}')
        return result

    def transform(self, *transformations):
        self.transformations.extend(transformations)
        return self

    def valid(self, result: Any) -> bool:
        return all(c(result) for c in self.constraints)


class FormulaParameter(Parameter):
    formulas = True
    pattern = '^.*$'
    illegal_patterns = tuple()
    type_ = str

    def __init__(self, name=None, default=None, optional=False, info=None, choices=None,
                 pattern=None, formulas=None):
        super().__init__(name, default, optional, info, choices)
        if pattern is not None:
            self.pattern = pattern
        if formulas is not None:
            self.formulas = formulas

    def load(self, specification):
        if self.pattern:
            specification = str(specification)
            match = re.match(self.pattern, specification)
            if match is None:
                raise ValueError(f'"{self.name}" Does not match pattern \'{self.pattern}\' '
                                 f'(\'{specification}\')')
            illegal = list(filter(
                lambda pattern: re.search(pattern, specification) is not None,
                self.illegal_patterns
            ))
            if illegal:
                match = re.search(illegal[0], specification)
                raise ValueError(f'"{self.name}" Contains illegal patterns '
                                 f'({repr(match.group())} forbidden by {repr(illegal[0])})')
        if self.formulas and isinstance(specification, str):
            specification = self.evaluate_formula(specification)
        return super().load(self.type_(specification))

    def evaluate_formula(self, formula):
        raise NotImplementedError


class Bool(FormulaParameter):
    pattern = '^[Tt]rue|[Ff]alse$'
    type_ = bool

    def evaluate_formula(self, formula):
        return formula.lower() == 'true'


class Integer(FormulaParameter):
    pattern = r'[()*+-/\s\d]+'
    illegal_patterns = ('[^/]/[^/]', '[a-zA-Z]', '{.*}', '\[.*\]')
    type_ = int

    def evaluate_formula(self, formula):
        return safe_math_eval(formula)


class Number(FormulaParameter):
    pattern = r'[()*+-/\s\da-zE.]+'
    illegal_patterns = Integer.illegal_patterns[2:]
    type_ = float
    np_members = ['sin', 'cos', 'tan', 'arcsin', 'arccos', 'arctan', 'hypot', 'arctan2',
                  'degrees', 'radians', 'unwrap', 'deg2rad', 'rad2deg', 'sinh', 'cosh', 'tanh',
                  'arcsinh', 'arccosh', 'arctanh', 'around', 'round_', 'rint', 'fix', 'floor',
                  'ceil', 'trunc', 'exp', 'expm1', 'exp2', 'log', 'log10', 'log2', 'log1p',
                  'logaddexp', 'logaddexp2', 'i0', 'sinc', 'sqrt', 'cbrt', 'square', 'absolute',
                  'fabs', 'sign', 'pi']

    def evaluate_formula(self, formula):
        scope = {m: getattr(np, m) for m in self.np_members} if np is not None else None
        return safe_math_eval(formula, locals_dict=scope)


class String(FormulaParameter):
    def evaluate_formula(self, formula):
        return formula


class PhysicalQuantity(FormulaParameter):
    pattern = r'[{}()*+-/\s\da-zA-Z.^]+\s?\[.+\]'
    type_ = float

    def __init__(self, name=None, default=None, optional=False, info=None, choices=None,
                 pattern=None, formulas=None, *, unit):
        super().__init__(name, default, optional, info, choices, pattern, formulas)
        self.unit = unit

    def evaluate_formula(self, formula):
        scope = {}
        if np is not None:
            scope = {m: getattr(np, m) for m in Number.np_members}
        if physical_constants is not None:
            scope.update({self.convert_constant_name(k): v[0] for k, v in physical_constants.items()})
            for key in physical_constants.keys():
                key_id = '{' + key + '}'
                if key_id in formula:
                    formula = formula.replace(key_id, self.convert_constant_name(key))
        m_units = re.search('\s*\[(.+)\]$', formula)
        unit = self.derive_unit(m_units.groups()[0])
        magnitude = safe_math_eval(formula.replace(m_units.group(), ''), locals_dict=scope)
        declared_unit = self.derive_unit(self.unit)
        return units.convert((magnitude, unit), declared_unit)

    @staticmethod
    def derive_unit(unit):
        if isinstance(unit, str):
            d_units = {u_id: units[u_id]
                       for u_id in re.findall(r'[a-zA-Z][a-zA-Z_0-9]*', unit)}
            return eval(unit.replace('^', '**'), {'__builtins__': None}, d_units)
        return unit

    @staticmethod
    def convert_constant_name(name):
        replacements = [('/', '_over_'), ('-', '_'), ('.', ''), ('(', '_'), (')', '_'), (',', ''),
                        ('{', ''), ('}', ''), (' ', '_')]
        return re.sub('__+', '_', reduce(lambda n, r: n.replace(*r), replacements, name))


class Characteristic:
    def __init__(self, parameter, func):
        self.parameter = parameter
        self.func = func

    def __eq__(self, val):
        self.parameter.constraints.append(Constraint(
            lambda x: self.func(x) == val,
            f'Not equal ({self.func.__name__}({{value}}) ≠ {val})'
        ))
        return self.parameter

    def __ge__(self, lim):
        self.parameter.constraints.append(Constraint(
            lambda x: self.func(x) >= lim,
            f'Too small ({self.func.__name__}({{value}}) < {lim})'
        ))
        return self.parameter

    def __gt__(self, lim):
        self.parameter.constraints.append(Constraint(
            lambda x: self.func(x) > lim,
            f'Too small ({self.func.__name__}({{value}}) ≤ {lim})'
        ))
        return self.parameter

    def __le__(self, lim):
        self.parameter.constraints.append(Constraint(
            lambda x: self.func(x) <= lim,
            f'Too large ({self.func.__name__}({{value}}) > {lim})'
        ))
        return self.parameter

    def __lt__(self, lim):
        self.parameter.constraints.append(Constraint(
            lambda x: self.func(x) < lim,
            f'Too large ({self.func.__name__}({{value}}) ≥ {lim})'
        ))
        return self.parameter

    def __ne__(self, val):
        self.parameter.constraints.append(Constraint(
            lambda x: self.func(x) != val,
            f'Must not be equal ({self.func.__name__}({{value}}) = {val})'
        ))
        return self.parameter


class Constraint:
    def __init__(self, func, error_msg):
        self.func = func
        self.error_msg = error_msg

    def __call__(self, val):
        if not self.func(val):
            if callable(self.error_msg):
                raise ValueError(self.error_msg(val))
            try:
                error_msg = self.error_msg.format(value=repr(val))
            except KeyError:
                error_msg = self.error_msg
            raise ValueError(error_msg)
        return True


class VectorFactory:
    def __getitem__(self, kind):
        return partial(Vector, kind=kind)

    @property
    def container(self):
        return Vector.container

    @container.setter
    def container(self, container):
        Vector.container = container


class Vector(Parameter):
    container = list

    def __init__(self, kind, name=None, default=None, optional=False, info=None, choices=None,
                 n=None, as_=None, **kwargs):
        super().__init__(name, default, optional, info, choices)
        self.template = kind(name=name, **kwargs)
        if isinstance(n, int):
            if n <= 0:
                raise ValueError('Number of elements must be greater than zero')
            _ = self.len == n
        if as_ is not None:
            self.container = as_

    def __abs__(self):
        raise NotImplementedError

    def __eq__(self, val):
        self.template.__eq__(val)
        return self

    def __ge__(self, lim):
        self.template.__ge__(lim)
        return self

    def __gt__(self, lim):
        self.template.__gt__(lim)
        return self

    def __le__(self, lim):
        self.template.__le__(lim)
        return self

    def __lt__(self, lim):
        self.template.__lt__(lim)
        return self

    def __ne__(self, val):
        self.template.__ne__(val)
        return self

    def load(self, value: list):
        result = []
        for i, x in enumerate(value):
            # Set name in order to have error messages displayed correctly.
            self.template.name = f'{self.name}[{i}]'
            result.append(self.template.load(x))
        self.template.name = None
        return super().load(self.container(result))

    def as_(self, container):
        self.container = container
        return self


class Wrapper:
    def __init__(self, name=None, info=None):
        self.info = info
        self.transformations = []
        self.name = name or '.'
        self.field_name = None

    def __set_name__(self, owner, name):
        self.field_name = name

    def __repr__(self):
        return f'{self.__class__.__name__}(name={repr(self.name)}, info={repr(self.info)})'

    def __rshift__(self, transformation):
        try:
            return self.transform(*transformation)
        except TypeError:
            return self.transform(transformation)

    def load(self, value):
        return reduce(lambda x, t: t(x), self.transformations, value)

    def transform(self, *transformations):
        self.transformations.extend(transformations)
        return self


class Group(Wrapper):
    container = dict

    def __init__(self, *members, name=None, info=None, as_=None):
        super().__init__(name, info)
        self.members = list(members)
        if as_ is not None:
            self.container = as_

    def __iadd__(self, parameter):
        self.add(parameter)
        return self

    def add(self, parameter):
        self.members.append(parameter)

    def as_(self, container):
        self.container = container
        return self

    def extend(self, other):
        new_members = [copy.copy(m) for m in other.members]
        for member in new_members:
            if other.name:
                member.name = f'{other.name}.{member.name}'
        self.members.extend(new_members)

    def load(self, value: ConfigTree):
        result = OrderedDict()
        for member in self.members:
            result[member.name] = member.load(value[member.name])
        if self.container in (list, tuple):
            result = self.container(result.values())
        else:
            result = self.container(**result)
        return super().load(result)


class FallbackGroup(Wrapper):
    def __init__(self, primary, proxies, name=None, info=None):
        super().__init__(name, info)
        self.primary = primary
        self.proxies = proxies

    def __set_name__(self, owner, name):
        super().__set_name__(owner, name)
        self.primary.__set_name__(owner, name)

    def __or__(self, parameter):
        return self.fallback(parameter)

    @property
    def optional(self):
        return self.primary.optional or any(map(lambda p: p.optional, self.proxies))

    @optional.setter
    def optional(self, x):
        pass

    def load(self, config):
        for parameter in [self.primary] + self.proxies:
            try:
                return parameter.load(config[parameter.name])
            except ConfigMissingException:
                continue
        for parameter in [self.primary] + self.proxies:
            if parameter.optional:
                return parameter.default
        raise ConfigMissingException(
            f'"{self.primary.name}" No fallback found '
            f'(checked the following parameters: {", ".join(map(lambda p: p.name, self.proxies))})'
        )

    def fallback(self, parameter):
        self.proxies.append(parameter)
        return self


class ComplementaryGroup(Wrapper):
    container = dict

    def __init__(self, *members, name=None, info=None, as_=None):
        super().__init__(name, info)
        members, rules = zip(*members)
        self.members = members
        self.rules = rules
        if as_ is not None:
            self.container = as_

    def __iadd__(self, parameter_and_rule):
        self.add(*parameter_and_rule)
        return self

    def add(self, parameter, rule):
        self.members.append(parameter)
        self.rules.append(rule)

    def as_(self, container):
        self.container = container
        return self

    def load(self, config):
        missing = list(filter(lambda p: p[1].name not in config, enumerate(self.members)))
        if len(missing) != 1:
            raise ValueError(f'"{self.name}" Too {"many" if not missing else "few"} '
                             f'parameters were specified ({abs(len(missing) - 1)})')

        members = [x for x in self.members]
        rules = [x for x in self.rules]
        complement = members.pop(missing[0][0])
        completion_rule = rules.pop(missing[0][0])

        result = OrderedDict()
        for member in members:
            result[member.name] = member.load(config[member.name])
        result[complement.name] = completion_rule(**result)

        if self.container in (list, tuple):
            result = self.container(result.values())
        else:
            result = self.container(**result)

        return super().load(result)



import re


def safe_math_eval(string, unsafe_error=ValueError, locals_dict=None):
    is_unsafe = [
        any(map(
            lambda x: x in string,
            ('__', '()', '[]', '{}', 'lambda', ',', ';', ':', '"', "'")
        )),
        # Allow a-z for numpy or math functions.
        re.match(r'^[.*/+\-0-9a-zA-Z\s()_]+$', string) is None
    ]
    if any(is_unsafe):
        raise unsafe_error(f'Evaluation of {repr(string)} is considered unsafe')
    return eval(string, {'__builtins__': None}, locals_dict or {})



from collections import namedtuple
from unittest import TestCase
import unittest

import numpy as np
from pyhocon import ConfigFactory
from pyhocon.exceptions import ConfigMissingException
from scipy.constants import physical_constants

from hanna import Integer, Number, PhysicalQuantity, String
from hanna import ComplementaryGroup, Group, Vector, Vectors
from hanna.configuration import Configurable
from hanna.parameters import FallbackGroup
from hanna.physics import units
from hanna.physics.units import NumericalUnits, Pint, Units


class TestGeneral(TestCase):
    def test_underscores(self):
        class Test(Configurable):
            _foo = String()
            _bar = Integer()

        self.assertEqual(Test._foo.name, 'foo')
        self.assertEqual(Test._bar.name, 'bar')


class TestSections(TestCase):
    def test_general(self):
        class Book(Configurable):
            title = String()
            year = Integer()

        self.assertEqual(Book.title.name, 'title')
        self.assertEqual(Book.year.name, 'year')

        title, year = 'some title', 2018
        book = Book(ConfigFactory.from_dict(dict(title=title, year=year)))
        self.assertEqual(book.title, title)
        self.assertEqual(book.year, year)

    def test_at_class_creation(self):
        class Book(Configurable, path='book'):
            title = String()
            year = Integer()

        self.assertEqual(Book.title.name, 'book.title')
        self.assertEqual(Book.year.name, 'book.year')

        title, year = 'some title', 2018
        book = Book(ConfigFactory.from_dict({
            'book': {
                'title': title,
                'year': year,
            }
        }))
        self.assertEqual(book.title, title)
        self.assertEqual(book.year, year)

        class Book(Configurable, path='library.book'):
            title = String()
            year = Integer()

        self.assertEqual(Book.title.name, 'library.book.title')
        self.assertEqual(Book.year.name, 'library.book.year')

        title, year = 'some title', 2018
        book = Book(ConfigFactory.from_dict({
            'library': {
                'book': {
                    'title': title,
                    'year': year,
                }
            }
        }))
        self.assertEqual(book.title, title)
        self.assertEqual(book.year, year)

    def test_as_part_of_the_name(self):
        class Book(Configurable):
            title = String('library.book.title')
            year = Integer('library.book.year')

        self.assertEqual(Book.title.name, 'library.book.title')
        self.assertEqual(Book.year.name, 'library.book.year')

        title, year = 'some title', 2018
        book = Book(ConfigFactory.from_dict({
            'library': {
                'book': {
                    'title': title,
                    'year': year,
                }
            }
        }))
        self.assertEqual(book.title, title)
        self.assertEqual(book.year, year)

    def test_in_place_of_the_name(self):
        class Book(Configurable):
            title = String('library.book.')
            year = Integer('library.book.')

        self.assertEqual(Book.title.name, 'library.book.title')
        self.assertEqual(Book.year.name, 'library.book.year')

        title, year = 'some title', 2018
        book = Book(ConfigFactory.from_dict({
            'library': {
                'book': {
                    'title': title,
                    'year': year,
                }
            }
        }))
        self.assertEqual(book.title, title)
        self.assertEqual(book.year, year)

        title_par = String('library.book.')
        self.assertEqual(title_par.name, 'library.book.')

        class Book(Configurable):
            title = title_par

        self.assertEqual(Book.title.name, 'library.book.title')

    def test_combinations(self):
        class Book(Configurable, path='library'):
            title = String('book.')
            year = Integer('book.year')

        self.assertEqual(Book.title.name, 'library.book.title')
        self.assertEqual(Book.year.name, 'library.book.year')

        title, year = 'some title', 2018
        book = Book(ConfigFactory.from_dict({
            'library': {
                'book': {
                    'title': title,
                    'year': year,
                }
            }
        }))
        self.assertEqual(book.title, title)
        self.assertEqual(book.year, year)


class TestTransformations(TestCase):
    def test_1(self):
        class Book(Configurable):
            n_words_1 = String('title').transform(str.split, len)
            n_words_2 = String('title').transform(lambda x: len(x.split()))
            century = Integer('year').transform(lambda x: x // 1000 + 1)

        self.assertEqual(len(Book.n_words_1.transformations), 2)
        self.assertEqual(len(Book.n_words_2.transformations), 1)
        self.assertEqual(len(Book.century.transformations), 1)

        book = Book(ConfigFactory.from_dict(dict(title='This is the title', year=2018)))
        self.assertEqual(book.n_words_1, 4)
        self.assertEqual(book.n_words_2, 4)
        self.assertEqual(book.century, 3)

        import math

        class Test(Configurable):
            p1 = Number('number').transform(int)
            p2 = Number('number').transform(math.floor)
            p3 = Number('number').transform(math.ceil)
            p4 = Integer().transform(lambda x: x ** 2)
            p5 = Integer().transform(lambda x: 'a' * x)

        test = Test(ConfigFactory.from_dict(dict(number=3.14, p4=5, p5=3)))
        self.assertEqual(test.p1, 3)
        self.assertEqual(test.p2, 3)
        self.assertEqual(test.p3, 4)
        self.assertEqual(test.p4, 25)
        self.assertEqual(test.p5, 'aaa')

    def test_2(self):
        class Test(Configurable):
            p1 = Integer('p').transform(lambda x: x**2, lambda x: x**3, lambda x: x**4)
            p2 = Integer('p').transform(lambda x: x**2).transform(lambda x: x**3).transform(lambda x: x**4)

        self.assertEqual(len(Test.p1.transformations), 3)
        self.assertEqual(len(Test.p2.transformations), 3)

        test = Test(ConfigFactory.from_dict(dict(p=2)))
        self.assertEqual(test.p1, 16777216)
        self.assertEqual(test.p2, 16777216)

    def test_syntactic_sugar(self):
        class Test(Configurable):
            p1 = Integer('p') >> (lambda x: x**2, lambda x: x**3, lambda x: x**4)
            p2 = Integer('p') >> (lambda x: x**2) >> (lambda x: x**3) >> (lambda x: x**4)

        self.assertEqual(len(Test.p1.transformations), 3)
        self.assertEqual(len(Test.p2.transformations), 3)

        test = Test(ConfigFactory.from_dict(dict(p=2)))
        self.assertEqual(test.p1, 16777216)
        self.assertEqual(test.p2, 16777216)


class TestConstraints(TestCase):
    def test_integer(self):
        class Test(Configurable):
            p1 = Integer() >= 1
            p2 = Integer() > 2
            p3 = Integer() <= 3
            p4 = Integer() < 4
            p5 = Integer() == 5  # Not really useful for built-in types but we'll test it anyway.
            p6 = Integer() != 6

        try:
            Test(ConfigFactory.from_dict(dict(p1=1, p2=3, p3=3, p4=3, p5=5, p6=5)))
        except ValueError:
            self.fail('Constraints not applied correctly')

        with self.assertRaises(ValueError) as context:
            Test(ConfigFactory.from_dict(dict(p1=0, p2=3, p3=3, p4=3, p5=5, p6=5)))
        self.assertEqual('"p1" Too small (0 < 1)', str(context.exception))

        with self.assertRaises(ValueError) as context:
            Test(ConfigFactory.from_dict(dict(p1=1, p2=2, p3=3, p4=3, p5=5, p6=5)))
        self.assertEqual('"p2" Too small (2 ≤ 2)', str(context.exception))

        with self.assertRaises(ValueError) as context:
            Test(ConfigFactory.from_dict(dict(p1=1, p2=3, p3=4, p4=3, p5=5, p6=5)))
        self.assertEqual('"p3" Too large (4 > 3)', str(context.exception))

        with self.assertRaises(ValueError) as context:
            Test(ConfigFactory.from_dict(dict(p1=1, p2=3, p3=3, p4=4, p5=5, p6=5)))
        self.assertEqual('"p4" Too large (4 ≥ 4)', str(context.exception))

        with self.assertRaises(ValueError) as context:
            Test(ConfigFactory.from_dict(dict(p1=1, p2=3, p3=3, p4=3, p5=6, p6=5)))
        self.assertEqual('"p5" Not equal (6 ≠ 5)', str(context.exception))

        with self.assertRaises(ValueError) as context:
            Test(ConfigFactory.from_dict(dict(p1=1, p2=3, p3=3, p4=3, p5=5, p6=6)))
        self.assertEqual('"p6" Must not be equal (6 = 6)', str(context.exception))

        class Test(Configurable):
            # noinspection PyTypeChecker
            p1 = 0 <= Integer() <= 2

        try:
            Test(ConfigFactory.from_dict(dict(p1=1)))
        except ValueError:
            self.fail('Constraints not applied correctly')

        with self.assertRaises(ValueError) as context:
            Test(ConfigFactory.from_dict(dict(p1=-1)))
        self.assertEqual('"p1" Too small (-1 < 0)', str(context.exception))

        with self.assertRaises(ValueError) as context:
            Test(ConfigFactory.from_dict(dict(p1=3)))
        self.assertEqual('"p1" Too large (3 > 2)', str(context.exception))

    def test_characteristics(self):
        class Test(Configurable):
            p1 = Integer().abs <= 5
            p2 = Integer().abs < 5
            p3 = Integer().abs >= 5
            p4 = Integer().abs > 5
            p5 = String().len <= 5
            p6 = String().len < 5
            p7 = String().len >= 5
            p8 = String().len > 5

        config = dict(
            p1=-5, p2=-4, p3=5, p4=6,
            p5='12345', p6='1234', p7='12345', p8='123456'
        )
        try:
            test = Test(ConfigFactory.from_dict(config))
        except ValueError:
            self.fail('Constraints not applied correctly')
        self.assertEqual(test.p1, config['p1'])
        self.assertEqual(test.p2, config['p2'])
        self.assertEqual(test.p3, config['p3'])
        self.assertEqual(test.p4, config['p4'])
        self.assertEqual(test.p5, config['p5'])
        self.assertEqual(test.p6, config['p6'])
        self.assertEqual(test.p7, config['p7'])
        self.assertEqual(test.p8, config['p8'])

    def test_custom_constraints(self):
        class Test(Configurable):
            p1 = Integer().constrain(lambda x: x <= 5, '')
            p2 = Integer().constrain(lambda x: x <= 5, 'Custom error message')
            p3 = Integer().constrain(lambda x: x <= 5, 'Wrong value: {value}')
            p4 = String().len > 5
            p5 = String().constrain(lambda x: len(x.split()) > 5, 'Not enough words: {value}')

        try:
            Test(ConfigFactory.from_dict(dict(p1=3, p2=3, p3=3, p4='123456', p5='1 2 3 4 5 6')))
        except ValueError:
            self.fail('Constraints not applied correctly')

        with self.assertRaises(ValueError) as context:
            Test(ConfigFactory.from_dict(dict(p1=8, p2=3, p3=3, p4='123456', p5='1 2 3 4 5 6')))
        self.assertEqual('"p1" ', str(context.exception))

        with self.assertRaises(ValueError) as context:
            Test(ConfigFactory.from_dict(dict(p1=3, p2=8, p3=3, p4='123456', p5='1 2 3 4 5 6')))
        self.assertEqual('"p2" Custom error message', str(context.exception))

        with self.assertRaises(ValueError) as context:
            Test(ConfigFactory.from_dict(dict(p1=3, p2=3, p3=8, p4='123456', p5='1 2 3 4 5 6')))
        self.assertEqual('"p3" Wrong value: 8', str(context.exception))

        with self.assertRaises(ValueError) as context:
            Test(ConfigFactory.from_dict(dict(p1=3, p2=3, p3=3, p4='12345', p5='1 2 3 4 5 6')))
        self.assertEqual('"p4" Too small (len(\'12345\') ≤ 5)', str(context.exception))

        with self.assertRaises(ValueError) as context:
            Test(ConfigFactory.from_dict(dict(p1=3, p2=3, p3=3, p4='123456', p5='1 2 3 4 5')))
        self.assertEqual('"p5" Not enough words: \'1 2 3 4 5\'', str(context.exception))


class TestVectors(TestCase):
    def test_general(self):
        class Test(Configurable):
            v = Vectors[Integer]()
        self.assertEqual(Test(ConfigFactory.from_dict(dict(v=[1, 2, 3]))).v, [1, 2, 3])
        Test.v = Vector(String, name='v')
        self.assertEqual(Test(ConfigFactory.from_dict(dict(v=['a', 'b', 'c']))).v, ['a', 'b', 'c'])

    def test_containers(self):
        class Test(Configurable):
            v1 = Vector(Integer, as_=tuple)
            v2 = Vector(Integer).as_(tuple)
            v3 = Vector(Integer, as_=set)
        test = Test(ConfigFactory.from_dict(dict(v1=[1, 2, 3], v2=[1, 2, 3], v3=[1, 1, 2, 2, 3])))
        self.assertEqual(test.v1, (1, 2, 3))
        self.assertEqual(test.v2, (1, 2, 3))
        self.assertEqual(test.v3, {1, 2, 3})

        Vector.container = tuple

        class Test(Configurable):
            v = Vector(Integer)
        self.assertEqual(Test(ConfigFactory.from_dict(dict(v=[1, 2, 3]))).v, (1, 2, 3))

        Vector.container = list

    def test_constraints(self):
        IntVector = Vectors[Integer]

        class Test(Configurable):
            v1 = IntVector() > 0

        with self.assertRaises(ValueError) as context:
            Test(ConfigFactory.from_dict(dict(v1=[-1, 1])))
        self.assertEqual('"v1[0]" Too small (-1 ≤ 0)', str(context.exception))

        Test.v1 = Test.v1 <= 4
        with self.assertRaises(ValueError) as context:
            Test(ConfigFactory.from_dict(dict(v1=[1, 2, 3, 4, 5, 6, 7, 8])))
        self.assertEqual('"v1[4]" Too large (5 > 4)', str(context.exception))

        Test.v1 = IntVector(name='v1', n=3)
        with self.assertRaises(ValueError) as context:
            Test(ConfigFactory.from_dict(dict(v1=[1, 2])))
        self.assertEqual('"v1" Not equal (len([1, 2]) ≠ 3)', str(context.exception))

        Test.v1 = IntVector(name='v1').len > 2
        with self.assertRaises(ValueError) as context:
            Test(ConfigFactory.from_dict(dict(v1=[1, 2])))
        self.assertEqual('"v1" Too small (len([1, 2]) ≤ 2)', str(context.exception))

    def test_tuple_from_multiplication(self):
        class Test(Configurable):
            v = 3 * Integer()

        with self.assertRaises(ValueError) as context:
            Test(ConfigFactory.from_dict(dict(v=[0, 1])))
        self.assertEqual('"v" Not equal (len((0, 1)) ≠ 3)', str(context.exception))
        with self.assertRaises(ValueError) as context:
            Test(ConfigFactory.from_dict(dict(v=[0, 1, 2, 3])))
        self.assertEqual('"v" Not equal (len((0, 1, 2, 3)) ≠ 3)', str(context.exception))

        self.assertEqual(Test(ConfigFactory.from_dict(dict(v=[0, 1, 2]))).v, (0, 1, 2))


class TestPatterns(TestCase):
    def test_strings(self):
        class Test(Configurable):
            p1 = String(pattern=r'Exactly \d+ times')

        with self.assertRaises(ValueError) as context:
            Test(ConfigFactory.from_dict(dict(p1='Exactly two times')))
        self.assertEqual(
            '"p1" Does not match pattern \'Exactly \\d+ times\' (\'Exactly two times\')',
            str(context.exception)
        )

        self.assertEqual(
            Test(ConfigFactory.from_dict(dict(p1='Exactly 2 times'))).p1,
            'Exactly 2 times'
        )

    def test_integers(self):
        class Test(Configurable):
            p1 = Integer()
        self.assertEqual(Test(ConfigFactory.from_dict(dict(p1='-1'))).p1, -1)
        Test.p1 = Integer('p1', pattern=r'\d+')
        with self.assertRaises(ValueError) as context:
            Test(ConfigFactory.from_dict(dict(p1='-1')))
        self.assertEqual('"p1" Does not match pattern \'\\d+\' (\'-1\')', str(context.exception))
        self.assertEqual(Test(ConfigFactory.from_dict(dict(p1='123'))).p1, 123)
        self.assertEqual(Test(ConfigFactory.from_dict(dict(p1=123))).p1, 123)

    def test_deactivating_patterns(self):
        class Test(Configurable):
            p1 = Integer()
        with self.assertRaises(ValueError):
            Test(ConfigFactory.from_dict(dict(p1='1 / 2')))

        class Test(Configurable):
            p1 = Integer(pattern=False)
        self.assertEqual(Test(ConfigFactory.from_dict(dict(p1='1 / 2'))).p1, 0)


class TestGroups(TestCase):
    def test_general(self):
        class Test(Configurable):
            g = Group(String('first'), String('second'))
        self.assertEqual(Test.g.field_name, 'g')
        self.assertEqual(
            Test(ConfigFactory.parse_string('first = "1"\nsecond = "2"')).g,
            {'first': '1', 'second': '2'}
        )

    def test_containers(self):
        dim_type = namedtuple('dimensions', ['width', 'height'])

        class Test(Configurable):
            g1 = Group(Integer('width'), Integer('height'), as_=dim_type)
            g2 = Group(Integer('width'), Integer('height')).as_(dim_type)

        test = Test(ConfigFactory.parse_string('width = 1\nheight = 2'))
        self.assertEqual(type(test.g1), dim_type)
        self.assertEqual(repr(test.g1), 'dimensions(width=1, height=2)')
        self.assertEqual(type(test.g2), dim_type)
        self.assertEqual(repr(test.g2), 'dimensions(width=1, height=2)')

        class Test(Configurable):
            g1 = Group(Integer('a'), Integer('b'), Integer('c'), as_=tuple)
        test = Test(ConfigFactory.parse_string('a = 1\nb = 2\nc = 3'))
        self.assertEqual(test.g1, (1, 2, 3))

    def test_transformations(self):
        class Rectangle(Configurable):
            area = Group(Integer('width'), Integer('height')).transform(
                lambda x: x['width'] * x['height']
            )
        self.assertEqual(Rectangle(ConfigFactory.parse_string('width = 2\nheight = 3')).area, 6)

        class Rectangle(Configurable):
            area = Group(Integer('width'), Integer('height'), as_=tuple).transform(
                lambda x: x[0] * x[1]
            )
        self.assertEqual(Rectangle(ConfigFactory.parse_string('width = 2\nheight = 3')).area, 6)

        class Rectangle(Configurable):
            area = Group(Integer('width'), Integer('height'), as_=tuple) >> (lambda x: x[0] * x[1])
        self.assertEqual(Rectangle(ConfigFactory.parse_string('width = 2\nheight = 3')).area, 6)

    def test_sections(self):
        class Test(Configurable):
            g1 = Group(String('foo'), name='first.group')
            g2 = Group(String('bar'), name='second.group')

        test = Test(ConfigFactory.parse_string('first { group { foo = "foo-spec"}}\n'
                                               'second { group { bar = "bar-spec"}}'))
        self.assertEqual(test.g1, {'foo': 'foo-spec'})
        self.assertEqual(test.g2, {'bar': 'bar-spec'})

    def test_extensions(self):
        class Test(Configurable):
            g = Group(Integer('a'))
        Test.g.add(Integer('b'))
        Test.g += Integer('c')
        Test.g.extend(Group(Integer('d'), name='other'))

        test = Test(ConfigFactory.parse_string(
            'a = 1\n'
            'b = 2\n'
            'c = 3\n'
            'other {\n d = 4 }\n'
        ))
        self.assertEqual(test.g, {'a': 1, 'b': 2, 'c': 3, 'other.d': 4})


class TestChoices(TestCase):
    def test_general(self):
        class Test(Configurable):
            i = Integer(choices=(1, 2, 3))

        with self.assertRaises(ValueError) as context:
            Test(ConfigFactory.parse_string('i = 0'))
        self.assertEqual('"i" Illegal choice (0 not in (1, 2, 3))', str(context.exception))
        self.assertEqual(Test(ConfigFactory.parse_string('i = 1')).i, 1)

    def test_extending(self):
        class Test(Configurable):
            i = Integer(choices=[1, 2])

        with self.assertRaises(ValueError) as context:
            Test(ConfigFactory.parse_string('i = 3'))
        self.assertEqual('"i" Illegal choice (3 not in [1, 2])', str(context.exception))
        Test.i.choices.append(3)
        self.assertEqual(Test(ConfigFactory.parse_string('i = 3')).i, 3)


class TestOptionalParameters(TestCase):
    def test_general(self):
        class Test(Configurable):
            p = Integer(default=1)
        self.assertTrue(Test.p.optional)
        self.assertEqual(Test(None).p, 1)
        self.assertEqual(Test(ConfigFactory.parse_string('missing { p = 1 }')).p, 1)

        class Test(Configurable):
            p1 = Integer(default=None)
            p2 = Integer(optional=True)
        self.assertFalse(Test.p1.optional)
        self.assertTrue(Test.p2.optional)

        with self.assertRaises(ConfigMissingException):
            Test(ConfigFactory.parse_string('p2 = 1'))
        self.assertEqual(Test(ConfigFactory.parse_string('p1 = 1')).p2, None)

        class Test(Configurable):
            p = Integer()
        self.assertFalse(Test.p.optional)
        with self.assertRaises(ConfigMissingException) as context:
            Test(ConfigFactory.parse_string('missing { p = 1 }'))
        self.assertEqual("'No configuration setting found for key p'", str(context.exception))


class TestFallbackGroups(TestCase):
    def test_general(self):
        class Test(Configurable):
            p1 = Integer().fallback(Integer('p2'))
        self.assertEqual(Test.p1.field_name, 'p1')
        self.assertEqual(Test(ConfigFactory.parse_string('p1 = 1')).p1, 1)
        self.assertEqual(Test(ConfigFactory.parse_string('p2 = 2')).p1, 2)

    def test_from_or(self):
        class Test(Configurable):
            p1 = Integer('p1') | Integer('p2') | Integer('p3')
        self.assertEqual(Test.p1.field_name, 'p1')
        self.assertEqual(Test(ConfigFactory.parse_string('p1 = 1')).p1, 1)
        self.assertEqual(Test(ConfigFactory.parse_string('p2 = 2')).p1, 2)
        self.assertEqual(Test(ConfigFactory.parse_string('p3 = 3')).p1, 3)

    def test_missing(self):
        class Test(Configurable):
            p1 = Integer('p1') | Integer('p2')
        self.assertIsInstance(Test.p1, FallbackGroup)
        self.assertEqual(Test.p1.field_name, 'p1')
        with self.assertRaises(ConfigMissingException):
            Test(ConfigFactory.parse_string('p3 = 3'))


class TestPhysicalParameters(TestCase):
    def test_general(self):
        class Test(Configurable):
            p = PhysicalQuantity(unit='mm')
        self.assertEqual(Test(ConfigFactory.parse_string('p = "1 [mm]"')).p, 1.)
        self.assertEqual(Test(ConfigFactory.parse_string('p = "1 [cm]"')).p, 10.)
        self.assertEqual(Test(ConfigFactory.parse_string('p = "1 [m]"')).p, 1000.)

    def test_alternative_declaration(self):
        class Test(Configurable):
            p = PhysicalQuantity(unit=units.mm)
        self.assertEqual(Test(ConfigFactory.parse_string('p = "1 [mm]"')).p, 1.)
        self.assertEqual(Test(ConfigFactory.parse_string('p = "1 [cm]"')).p, 10.)
        self.assertEqual(Test(ConfigFactory.parse_string('p = "1 [m]"')).p, 1000.)

        class Test(Configurable):
            p = PhysicalQuantity(unit=units['mm'])
        self.assertEqual(Test(ConfigFactory.parse_string('p = "1 [mm]"')).p, 1.)
        self.assertEqual(Test(ConfigFactory.parse_string('p = "1 [cm]"')).p, 10.)
        self.assertEqual(Test(ConfigFactory.parse_string('p = "1 [m]"')).p, 1000.)

    def test_changing_units(self):
        class Test(Configurable):
            p = PhysicalQuantity(unit=units.m)
        Test.p.unit = units.mm
        self.assertEqual(Test(ConfigFactory.parse_string('p = "1 [mm]"')).p, 1.)
        self.assertEqual(Test(ConfigFactory.parse_string('p = "1 [cm]"')).p, 10.)
        self.assertEqual(Test(ConfigFactory.parse_string('p = "1 [m]"')).p, 1000.)
        Test.p.unit = 'mm'
        self.assertEqual(Test(ConfigFactory.parse_string('p = "1 [mm]"')).p, 1.)
        self.assertEqual(Test(ConfigFactory.parse_string('p = "1 [cm]"')).p, 10.)
        self.assertEqual(Test(ConfigFactory.parse_string('p = "1 [m]"')).p, 1000.)

    def test_numericalunits_engine(self):
        units.engine = NumericalUnits
        self.assertIsInstance(units.engine, NumericalUnits)

        class Test(Configurable):
            p = PhysicalQuantity(unit='mm')
        # We need to use `self.assertAlmostEqual` here because the numericalunits package randomly
        # initializes each dimension with a numerical value and scales the corresponding units
        # with the required factors. Therefore if we do for example the conversion m -> mm then
        # what happens in fact is `(x * 1000) / x` with `x` a random number. Note the parentheses
        # which indicate that scaling of the single units happens first and conversion later on.
        # For some numbers `x` the conversion might not exactly yield the initial scaling factor.
        self.assertEqual(Test(ConfigFactory.parse_string('p = "1 [mm]"')).p, 1.)
        self.assertAlmostEqual(Test(ConfigFactory.parse_string('p = "1 [cm]"')).p, 10., delta=2e-14)
        self.assertAlmostEqual(Test(ConfigFactory.parse_string('p = "1 [m]"')).p, 1000., delta=2e-12)

        class Test(Configurable):
            p = PhysicalQuantity(unit=units.mm)
        self.assertEqual(Test(ConfigFactory.parse_string('p = "1 [mm]"')).p, 1.)
        self.assertAlmostEqual(Test(ConfigFactory.parse_string('p = "1 [cm]"')).p, 10., delta=2e-14)
        self.assertAlmostEqual(Test(ConfigFactory.parse_string('p = "1 [m]"')).p, 1000., delta=2e-12)

        units.engine = Pint

    def test_units_engine(self):
        units.engine = Units
        self.assertIsInstance(units.engine, Units)

        class Test(Configurable):
            p = PhysicalQuantity(unit='mm')
        self.assertEqual(Test(ConfigFactory.parse_string('p = "1 [mm]"')).p, 1.)
        self.assertEqual(Test(ConfigFactory.parse_string('p = "1 [cm]"')).p, 10.)
        self.assertEqual(Test(ConfigFactory.parse_string('p = "1 [m]"')).p, 1000.)

        class Test(Configurable):
            p = PhysicalQuantity(unit=units.mm)
        self.assertEqual(Test(ConfigFactory.parse_string('p = "1 [mm]"')).p, 1.)
        self.assertEqual(Test(ConfigFactory.parse_string('p = "1 [cm]"')).p, 10.)
        self.assertEqual(Test(ConfigFactory.parse_string('p = "1 [m]"')).p, 1000.)

        units.engine = Pint


class TestFormulas(TestCase):
    def test_integers(self):
        class Test(Configurable):
            p = Integer()
        self.assertEqual(Test(ConfigFactory.parse_string('p = "1 + 2"')).p, 3)
        self.assertEqual(Test(ConfigFactory.parse_string('p = "2*3"')).p, 6)
        self.assertEqual(Test(ConfigFactory.parse_string('p = "2 ** 3"')).p, 8)
        self.assertEqual(Test(ConfigFactory.parse_string('p = "(1 + 2)**3"')).p, 27)
        self.assertEqual(Test(ConfigFactory.parse_string('p = "16 // 2"')).p, 8)
        with self.assertRaises(ValueError):
            Test(ConfigFactory.parse_string('p = "16 / 2"'))
        with self.assertRaises(ValueError):
            Test(ConfigFactory.parse_string('p = "2 * pi"'))
        with self.assertRaises(ValueError):
            Test(ConfigFactory.parse_string('p = "[1, 2, 3]"'))

    def test_numbers(self):
        class Test(Configurable):
            p = Number()
        self.assertEqual(Test(ConfigFactory.parse_string('p = "1 + 2"')).p, 3.)
        self.assertEqual(Test(ConfigFactory.parse_string('p = "2*3"')).p, 6.)
        self.assertEqual(Test(ConfigFactory.parse_string('p = "2 ** 3"')).p, 8.)
        self.assertEqual(Test(ConfigFactory.parse_string('p = "(1 + 2)**3"')).p, 27.)
        self.assertEqual(Test(ConfigFactory.parse_string('p = "16 / 2"')).p, 8.)
        self.assertEqual(Test(ConfigFactory.parse_string('p = "2 * pi"')).p, 2. * np.pi)
        self.assertEqual(Test(ConfigFactory.parse_string('p = "2 * sin(pi/2)"')).p,
                         2. * np.sin(np.pi/2))

    def test_physical_quantities(self):
        class Test(Configurable):
            p = PhysicalQuantity(unit='kg')
        self.assertEqual(
            Test(ConfigFactory.parse_string('p = "2 * {electron mass} [kg]"')).p,
            2. * physical_constants['electron mass'][0]
        )

        class Test(Configurable):
            p = PhysicalQuantity(unit='eV / K')
        self.assertEqual(
            Test(ConfigFactory.parse_string('p = "{Boltzmann constant in eV/K} [eV/K]"')).p,
            physical_constants['Boltzmann constant in eV/K'][0]
        )

        class Test(Configurable):
            p = PhysicalQuantity(unit='1 / m**3')
        self.assertEqual(
            Test(ConfigFactory.parse_string('p = "{Loschmidt constant (273.15 K, 100 kPa)} [1/m^3]"')).p,
            physical_constants['Loschmidt constant (273.15 K, 100 kPa)'][0]
        )

    def test_deactivating_formulas(self):
        class Test(Configurable):
            p = PhysicalQuantity(unit='kg', formulas=True)
        self.assertEqual(
            Test(ConfigFactory.parse_string('p = "{electron mass} [kg]"')).p,
            physical_constants['electron mass'][0]
        )

        class Test(Configurable):
            p = PhysicalQuantity(unit='kg', formulas=False)
        with self.assertRaises(ValueError):
            Test(ConfigFactory.parse_string('p = "{electron mass} [kg]"'))


class TestComplementaryGroups(TestCase):
    def test_general(self):
        class Test(Configurable):
            g = ComplementaryGroup(
                (Integer('width'), lambda height, area: area // height),
                (Integer('height'), lambda width, area: area // width),
                (Integer('area'), lambda width, height: width * height)
            )
        self.assertEqual(Test.g.field_name, 'g')
        self.assertEqual(
            Test(ConfigFactory.parse_string('width = 2\nheight=3')).g,
            {'width': 2, 'height': 3, 'area': 6}
        )
        self.assertEqual(
            Test(ConfigFactory.parse_string('width = 2\narea=6')).g,
            {'width': 2, 'height': 3, 'area': 6}
        )
        self.assertEqual(
            Test(ConfigFactory.parse_string('area = 6\nheight=3')).g,
            {'width': 2, 'height': 3, 'area': 6}
        )


class TestAutocompleteParameterNames(TestCase):
    @classmethod
    def setUpClass(cls):
        from hanna.addons import AutocompleteParameterNames
        AutocompleteParameterNames.install()
        cls.AddOn = AutocompleteParameterNames

    @classmethod
    def tearDownClass(cls):
        cls.AddOn.uninstall()

    def test_general(self):
        from hanna import Configurable, Integer, String

        class Test(Configurable):
            foo = String()
        i = Integer()
        Test.bar = i
        self.assertEqual(Test.foo.name, 'foo')
        self.assertEqual(Test.bar.name, 'bar')
        self.assertEqual(i.name, None)

    def test_install_uninstall(self):
        self.AddOn.uninstall()
        from hanna import Configurable, String

        class Test(Configurable):
            pass
        Test.foo = String()
        self.assertEqual(Test.foo.name, None)

        self.AddOn.install()
        from hanna import Configurable, String

        class Test(Configurable):
            pass
        Test.bar = String()
        self.assertEqual(Test.bar.name, 'bar')

    def test_class_wide_sections(self):
        from hanna import Configurable, Integer, String

        class Test(Configurable, path='section'):
            foo = String()
        Test.bar = Integer()
        self.assertEqual(Test.foo.name, 'section.foo')
        self.assertEqual(Test.bar.name, 'section.bar')


if __name__ == '__main__':
    unittest.main()



from .configuration import Configurable
from .parameters import Bool, Integer, Number, String, PhysicalQuantity, \
    ComplementaryGroup, Group, Vector, VectorFactory

Vectors = VectorFactory()
del VectorFactory



# Set a backend for physical units, e.g.:
#    * numericalunits (https://pypi.org/project/numericalunits/)
#    * pint (https://pint.readthedocs.io/en/latest/)
#    * units (https://pypi.org/project/units/)

try:
    import pint
except ImportError:
    pint = None

try:
    import numericalunits
except ImportError:
    numericalunits = None

try:
    import units
except ImportError:
    units = None


class Engine:
    backend = None
    backend_url = None

    def __init__(self):
        if self.backend is None:
            raise NotImplementedError(f'{self.__class__.__name__} engine is not available. '
                                      f'Please install the {self.backend_url} package first.')

    def __getitem__(self, item):
        raise NotImplementedError

    def convert(self, quantity, to_unit):
        raise NotImplementedError


class NumericalUnits(Engine):
    backend = numericalunits
    backend_url = 'https://pypi.org/project/numericalunits/'

    def __init__(self):
        super().__init__()

    def __getitem__(self, item):
        return getattr(self.backend, item)

    def convert(self, quantity, to_unit):
        return quantity[0] * (quantity[1] / to_unit)


class Pint(Engine):
    backend = pint
    backend_url = 'https://pypi.org/project/Pint/'

    def __init__(self):
        super().__init__()
        self._ur = self.backend.UnitRegistry()

    def __getitem__(self, item):
        return getattr(self._ur, item)

    def convert(self, quantity, to_unit):
        return (quantity[0] * quantity[1]).to(to_unit).magnitude


class Units(Engine):
    backend = units
    backend_url = 'https://pypi.org/project/units/'

    def __init__(self):
        super().__init__()
        from units.predefined import define_units
        define_units()

    def __getitem__(self, item):
        return units.unit(item)

    def convert(self, quantity, to_unit):
        return to_unit(quantity[1](quantity[0])).num


class UnitsProxy:
    def __init__(self):
        self._engine = None

    def __getattr__(self, item):
        if isinstance(item, str):
            return self._engine[item]
        return item

    def __getitem__(self, item):
        if isinstance(item, str):
            return self._engine[item]
        return item

    @property
    def engine(self):
        return self._engine

    @engine.setter
    def engine(self, cls):
        if not issubclass(cls, Engine):
            raise TypeError(f'Units engine must be a subclass of {Engine.__name__}')
        self._engine = cls()

    def convert(self, quantity, to_unit):
        """
        Convert the given quantity to different units.
        
        quantity : tuple, length 2
            First entry is the magnitude, second entry is the unit.
        to_unit : unit
            The unit to which the quantity is converted.
            
        Returns
        -------
        converted : float
            The magnitude of the converted quantity.
        """
        return self.engine.convert(quantity, to_unit)



from setuptools import setup


def readme():
    with open('README.rst') as f:
        return f.read()


def version():
    with open('hanna/VERSION') as f:
        return f.read()

setup(
    name='hanna',
    version=version(),
    description='Turns your configurations into actions',
    long_description=readme(),
    long_description_content_type='text/x-rst',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development',
    ],
    keywords=['configuration', 'framework'],
    url='https://gitlab.com/Dominik1123/Hanna',
    author='Dominik Vilsmeier',
    author_email='dominik.vilsmeier1123@gmail.com',
    license='BSD-3-Clause',
    packages=[
        'hanna',
        'hanna.physics',
    ],
    install_requires=[
        'pyhocon',
    ],
    python_requires='>=3.6',
    include_package_data=True,
    zip_safe=False
)



import copy

import hanna
from . import configuration
from .parameters import Parameter, Wrapper


class AddOn:
    @classmethod
    def install(cls):
        pass

    @classmethod
    def uninstall(cls):
        pass


class AutocompleteParameterNames(AddOn):
    """
    This add-on enables autocomplete of parameter names when being set on a `Configurable` subclass
    outside of a class definition.
    
    .. warning::
       As a consequence of installing this add-on the `Configurable` class will receive a specific
       metaclass which handles the autocomplete. In case you use custom metaclass as well this
       might lead to conflicts.
       
    .. note::
       As a consequence of installing this add-on each subclass of `Configurable` which specifies
       a `path` during class definition will receive an additional class attribute `_config_path`
       which holds the value of the specified path.
    """

    Configurable = configuration.Configurable

    class AutoCompleter(type):
        @classmethod
        def __prepare__(mcs, name, bases, **kwargs):
            namespace = super().__prepare__(name, bases)
            if 'path' in kwargs:
                namespace['_config_path'] = kwargs.get('path')
            return namespace

        def __setattr__(self, key, value):
            if isinstance(value, (Parameter, Wrapper)):
                value = copy.deepcopy(value)
                value.__set_name__(self, key)
                config_path = getattr(self, '_config_path', '')
                if config_path:
                    value.name = f'{config_path}.{value.name}'
            super().__setattr__(key, value)

    @classmethod
    def install(cls):
        super().install()
        configuration.Configurable = cls.AutoCompleter(
            configuration.Configurable.__name__,
            configuration.Configurable.__bases__,
            dict(**configuration.Configurable.__dict__)
        )
        hanna.Configurable = configuration.Configurable

    @classmethod
    def uninstall(cls):
        super().uninstall()
        configuration.Configurable = cls.Configurable
        hanna.Configurable = cls.Configurable



import warnings
from .units import UnitsProxy, Pint

units = UnitsProxy()
try:
    units.engine = Pint
except NotImplementedError:
    warnings.warn('The default units engine is not available. You need to set the engine manually '
                  'via `from hanna.physics import units; units.engine = ...`.')
del UnitsProxy
